# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os
import re
from collections import defaultdict

MANIFEST_NAMES = ("__manifest__.py", "__openerp__.py")


# Copy-pasted from OCA/maintainer-tools
def get_manifest_path(addon_dir):
    for manifest_name in MANIFEST_NAMES:
        manifest_path = os.path.join(addon_dir, manifest_name)
        if os.path.isfile(manifest_path):
            return manifest_path


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[39m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    ENDD = "\033[22m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def clean_text(text):
    """Clean text by removing patterns like '13.0', '[13.0]' or '[IMP]'."""
    return re.sub(r"\[.*\]|\d+\.\d+", "", text).strip()


def defaultdict_from_dict(d):
    nd = lambda: defaultdict(nd)  # noqa
    ni = nd()
    ni.update(d)
    return ni



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os

import requests

GITHUB_API_URL = "https://api.github.com"


def request(url, method="get", params=None, json=None):
    """Request GitHub API."""
    headers = {"Accept": "application/vnd.github.groot-preview+json"}
    if os.environ.get("GITHUB_TOKEN"):
        token = os.environ.get("GITHUB_TOKEN")
        headers.update({"Authorization": f"token {token}"})
    full_url = "/".join([GITHUB_API_URL, url])
    kwargs = {"headers": headers}
    if json:
        kwargs.update(json=json)
    if params:
        kwargs.update(params=params)
    response = getattr(requests, method)(full_url, **kwargs)
    if not response.ok:
        raise RuntimeError(response.text)
    return response.json()


def get_original_pr(upstream_org: str, repo_name: str, branch: str, commit_sha: str):
    """Return original GitHub PR data of a commit."""
    gh_commit_pulls = request(
        f"repos/{upstream_org}/{repo_name}/commits/{commit_sha}/pulls"
    )
    gh_commit_pull = [
        data
        for data in gh_commit_pulls
        if (
            data["base"]["ref"] == branch
            and data["base"]["repo"]["full_name"] == f"{upstream_org}/{repo_name}"
        )
    ]
    return gh_commit_pull and gh_commit_pull[0] or {}



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import contextlib
import re
import subprocess
from collections import abc

import git as g

from . import misc
from .misc import bcolors as bc

PO_FILE_REGEX = re.compile(r".*i18n/.+\.pot?$")


class Branch:
    def __init__(self, repo, name, default_remote=None, check_remote=True):
        self.repo = repo
        if len(name.split("/", 1)) > 1:
            remote, name = name.split("/", 1)
        else:
            remote = default_remote
        self.name = name
        self.remote = None
        if remote:
            if check_remote and remote not in repo.remotes:
                raise ValueError(repo, remote)
            self.remote = remote

    def ref(self):
        ref = self.name
        if self.remote:
            ref = f"{self.remote}/{self.name}"
        return ref


class CommitPath(str):
    """Helper class to know if a base path is a directory or a file."""

    def __new__(cls, value):
        new_value = value.split("/", maxsplit=1)[0]
        obj = super().__new__(cls, new_value)
        obj.isdir = "/" in value
        return obj


class Commit:
    # Attributes used to check equality between commits.
    # We do not want to use the SHA here as it changed from one branch to another
    # when a commit is ported (obviously).
    base_equality_attrs = (
        "author_name",
        "author_email",
        "authored_datetime",
        "message",
    )
    other_equality_attrs = ("paths",)
    eq_strict = True

    def __init__(self, commit):
        """Initializes a new Commit instance from a GitPython Commit object."""
        self.raw_commit = commit
        self.author_name = commit.author.name
        self.author_email = commit.author.email
        self.authored_datetime = commit.authored_datetime.replace(
            tzinfo=None
        ).isoformat()
        self.summary = commit.summary
        self.message = commit.message
        self.hexsha = commit.hexsha
        self.committed_datetime = commit.committed_datetime.replace(tzinfo=None)
        self.parents = [parent.hexsha for parent in commit.parents]
        self.files = {f for f in set(commit.stats.files.keys()) if "=>" not in f}
        self.paths = {CommitPath(f) for f in self.files}
        self.ported_commits = []

    def _get_equality_attrs(self):
        return [attr for attr in self.base_equality_attrs if hasattr(self, attr)] + [
            attr
            for attr in self.other_equality_attrs
            if self.__class__.eq_strict and hasattr(self, attr)
        ]

    def _lazy_eq_message(self, other):
        """Compare commit messages."""
        # If the subject has been put on two lines, 'git-am' won't preserve it
        # if '--keep-cr' option is not set, this generates false-positive.
        # Replace all carriage returns and double spaces by one space character
        # when performing the comparison.
        self_value = self.message.replace("\n", " ").replace("  ", " ")
        other_value = other.message.replace("\n", " ").replace("  ", " ")
        # 'git am' without '--keep' option removes text in '[]' brackets
        # generating false-positive.
        return misc.clean_text(self_value) == misc.clean_text(other_value)

    def __eq__(self, other):
        """Consider a commit equal to another if some of its keys are the same."""
        if not isinstance(other, Commit):
            return super().__eq__(other)
        if self.__class__.eq_strict:
            return all(
                [
                    getattr(self, attr) == getattr(other, attr)
                    for attr in self._get_equality_attrs()
                ]
            )
        else:
            checks = [
                (
                    self._lazy_eq_message(other)
                    if attr == "message"
                    else getattr(self, attr) == getattr(other, attr)
                )
                for attr in self._get_equality_attrs()
            ]
            return all(checks)

    def __repr__(self):
        attrs = ", ".join([f"{k}={v}" for k, v in self.__dict__.items()])
        return f"{self.__class__.__name__}({attrs})"

    @property
    def addons_created(self):
        """Returns the list of addons created by this commit."""
        addons = set()
        for diff in self.diffs:
            if (
                any(manifest in diff.b_path for manifest in misc.MANIFEST_NAMES)
                and diff.change_type == "A"
            ):
                addons.add(diff.b_path.split("/", maxsplit=1)[0])
        return addons

    @property
    def paths_to_port(self):
        """Return the list of file paths to port."""
        current_paths = {
            diff.a_path
            for diff in self.diffs
            if self._keep_diff_path(diff, diff.a_path)
        }.union(
            {
                diff.b_path
                for diff in self.diffs
                if self._keep_diff_path(diff, diff.b_path)
            }
        )
        ported_paths = set()
        for ported_commit in self.ported_commits:
            for diff in ported_commit.diffs:
                ported_paths.add(diff.a_path)
                ported_paths.add(diff.b_path)
        return current_paths - ported_paths

    def _keep_diff_path(self, diff, path):
        """Check if a file path should be ported."""
        # Ignore 'setup' files
        if path.startswith("setup"):
            return False
        # Ignore changes on po/pot files
        if PO_FILE_REGEX.match(path):
            return False
        return True

    @property
    def diffs(self):
        if self.raw_commit.parents:
            return self.raw_commit.diff(self.raw_commit.parents[0], R=True)
        return self.raw_commit.diff(g.NULL_TREE)


@contextlib.contextmanager
def no_strict_commit_equality():
    try:
        Commit.eq_strict = False
        yield
    finally:
        Commit.eq_strict = True


class PullRequest(abc.Hashable):
    eq_attrs = ("number", "url", "author", "title", "body", "merged_at")

    def __init__(
        self,
        number,
        url,
        author,
        title,
        body,
        merged_at,
        commits=None,
        paths=None,
        ported_paths=None,
    ):
        self.number = number
        self.url = url
        self.author = author
        self.title = title
        self.body = body
        self.merged_at = merged_at
        self.commits = commits if commits else []
        self.paths = set(paths) if paths else set()
        self.ported_paths = set(ported_paths) if ported_paths else set()

    def __eq__(self, other):
        if not isinstance(other, PullRequest):
            return super().__eq__(other)
        return all(
            [
                getattr(self, attr) == getattr(other, attr)
                for attr in self.__class__.eq_attrs
            ]
        )

    def __hash__(self):
        attr_values = tuple(getattr(self, attr) for attr in self.eq_attrs)
        return hash(attr_values)

    @property
    def paths_not_ported(self):
        return list(self.paths - self.ported_paths)


def run_pre_commit(repo, addon, commit=True, hook=None):
    # Run pre-commit
    print(f"\tRun {bc.BOLD}pre-commit{bc.END} and commit changes if any...")
    # First ensure that 'pre-commit' is initialized for the repository,
    # then run it (without checking the return code on purpose)
    subprocess.check_call("pre-commit install", shell=True)
    if hook:
        subprocess.run(f"pre-commit run {hook}", shell=True)
    else:
        subprocess.run("pre-commit run -a", shell=True)
    if repo.untracked_files or repo.is_dirty():
        repo.git.add("-A")
        if commit:
            repo.git.commit(
                "-m", f"[IMP] {addon}: black, isort, prettier", "--no-verify"
            )


def get_changed_paths(repo, modified=True, staged=True):
    """Return a list of file paths that have been changed.

    :param modified: include modified files (not added to the index)
    :param staged: include staged files (added to the index)
    :return: list of changed file paths
    """
    changed_diff = []
    if modified:
        changed_diff.extend(repo.index.diff(None))
    if staged:
        changed_diff.extend(repo.index.diff("HEAD"))
    return [diff.a_path or diff.b_path for diff in changed_diff]



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import json
import logging
import os
import pathlib
import shutil
from collections import defaultdict

from . import git as g, misc

_logger = logging.getLogger(__name__)


class UserCacheFactory:
    """User's cache manager factory."""

    def __init__(
        self,
        upstream_org: str,
        repo_name: str,
        addon: str,
        from_branch: g.Branch,
        to_branch: g.Branch,
        no_cache: bool,
    ):
        self._upstream_org = upstream_org
        self._repo_name = repo_name
        self._addon = addon
        self._from_branch = from_branch
        self._to_branch = to_branch
        self._no_cache = no_cache

    def build(self):
        """Build the cache manager."""
        if self._no_cache:
            return NoCache()
        try:
            cache = UserCache(
                self._upstream_org,
                self._repo_name,
                self._addon,
                self._from_branch,
                self._to_branch,
            )
        except Exception:
            # If the cache can't be used (whatever the reason) we fallback on a
            # fake-cache manager.
            _logger.warning(
                "No cache will be used: "
                "unable to initialize the cache folder in %s.",
                UserCache._get_dir_path(),
            )
            cache = NoCache()
        return cache


class NoCache:
    """Fake cache manager class.

    Used if the cache can't be used, e.g. no write access to cache folder.
    """

    def __init__(self, *args, **kwargs):
        """Initialize a fake user's cache manager."""

    def mark_commit_as_ported(self, commit_sha):
        # Do nothing
        pass

    def is_commit_ported(self, commit_sha):
        # A commit is always considered as not ported
        return False

    def store_commit_pr(self, commit_sha: str, data):
        # Do nothing
        pass

    def get_pr_from_commit(self, commit_sha: str):
        # No PR data to return
        return {}

    def clear(self):
        # Do nothing
        pass


class UserCache:
    """Manage the user's cache, in respect to XDG conventions.

    This class manages the following data:
        - a list of already ported commits from one branch to another.

    It allows to speed up further commit scans on a given module.
    """

    _cache_dirname = "oca-port"
    _ported_dirname = "ported"
    _to_port_dirname = "to_port"

    def __init__(
        self,
        upstream_org: str,
        repo_name: str,
        addon: str,
        from_branch: g.Branch,
        to_branch: g.Branch,
    ):
        """Initialize user's cache manager."""
        self._upstream_org = upstream_org
        self._repo_name = repo_name
        self._addon = addon
        self._from_branch = from_branch
        self._to_branch = to_branch
        self.dir_path = self._get_dir_path()
        self._ported_commits_path = self._get_ported_commits_path()
        self._ported_commits = self._get_ported_commits()
        self._commits_to_port_path = self._get_commits_to_port_path()
        self._commits_to_port = self._get_commits_to_port()

    @classmethod
    def _get_dir_path(cls):
        """Return the path of the cache directory."""
        default_cache_dir_path = pathlib.Path.home().joinpath(".cache")
        return pathlib.Path(
            os.environ.get("XDG_CACHE_HOME", default_cache_dir_path), cls._cache_dirname
        )

    def _get_ported_commits_path(self):
        """Return the file path storing ported commit."""
        file_name = (
            f"{self._addon}_{self._from_branch.name}_" f"to_{self._to_branch.name}.list"
        )
        return self.dir_path.joinpath(
            self._ported_dirname, self._upstream_org, self._repo_name, file_name
        )

    def _get_commits_to_port_path(self):
        """Return the file path storing cached data of commits to port."""
        file_name = (
            f"{self._addon}_{self._from_branch.name}_" f"to_{self._to_branch.name}.json"
        )
        return self.dir_path.joinpath(
            self._to_port_dirname, self._upstream_org, self._repo_name, file_name
        )

    def _get_ported_commits(self):
        self._ported_commits_path.parent.mkdir(parents=True, exist_ok=True)
        self._ported_commits_path.touch(exist_ok=True)
        return self._ported_commits_path.read_text().splitlines()

    def _get_commits_to_port(self):
        self._commits_to_port_path.parent.mkdir(parents=True, exist_ok=True)
        self._commits_to_port_path.touch(exist_ok=True)
        try:
            with self._commits_to_port_path.open() as file_:
                return json.load(file_, object_hook=misc.defaultdict_from_dict)
        except json.JSONDecodeError:
            # Mainly to handle empty files (first initialization of the cache)
            # but also to not crash if JSON files get corrupted.
            # Returns a "nested dict" object to not worry about checking keys
            nested_dict = lambda: defaultdict(nested_dict)  # noqa
            return nested_dict()

    def mark_commit_as_ported(self, commit_sha: str):
        """Mark commit as ported."""
        if self.is_commit_ported(commit_sha):
            return
        self._ported_commits.append(commit_sha)
        with self._ported_commits_path.open(mode="a") as file_:
            file_.write(f"{commit_sha}\n")

    def is_commit_ported(self, commit_sha: str):
        """Return `True` if commit is already ported."""
        return commit_sha in self._ported_commits

    def store_commit_pr(self, commit_sha: str, data):
        """Store the original PR data of a commit."""
        pr_number = data["number"]
        self._commits_to_port["pull_requests"][str(pr_number)] = data
        self._commits_to_port["commits"][commit_sha]["pr"] = pr_number
        try:
            with self._commits_to_port_path.open(mode="w") as file_:
                json.dump(self._commits_to_port, file_, indent=2)
        except Exception:
            pass

    def get_pr_from_commit(self, commit_sha: str):
        """Return the original PR data of a commit."""
        pr_number = self._commits_to_port["commits"][commit_sha]["pr"]
        if pr_number:
            return self._commits_to_port["pull_requests"][str(pr_number)]
        return {}

    def clear(self):
        """Clear the cache by removing the content of the cache directory."""
        if self._cache_dirname and str(self.dir_path).endswith(self._cache_dirname):
            shutil.rmtree(self.dir_path)



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os
import tempfile
import urllib.parse

import click

from .port_addon_pr import PortAddonPullRequest
from .utils import git as g
from .utils.misc import bcolors as bc

MIG_BRANCH_NAME = "{branch}-mig-{addon}"
MIG_MERGE_COMMITS_URL = (
    "https://github.com/OCA/maintainer-tools/wiki/Merge-commits-in-pull-requests"
)
MIG_TASKS_URL = (
    "https://github.com/OCA/maintainer-tools/wiki/Migration-to-version-{branch}"
    "#tasks-to-do-in-the-migration"
)
MIG_NEW_PR_TITLE = "[{to_branch}][MIG] {addon}"
MIG_NEW_PR_URL = (
    "https://github.com/{upstream_org}/{repo_name}/compare/"
    "{to_branch}...{user_org}:{mig_branch}?expand=1&title={title}"
)
MIG_TIPS = "\n".join(
    [
        f"\n{bc.BOLD}{bc.OKCYAN}The next steps are:{bc.END}",
        ("\t1) Reduce the number of commits " f"('{bc.DIM}OCA Transbot...{bc.END}'):"),
        f"\t\t=> {bc.BOLD}{MIG_MERGE_COMMITS_URL}{bc.END}",
        "\t2) Adapt the module to the {to_branch} version:",
        f"\t\t=> {bc.BOLD}" "{mig_tasks_url}" f"{bc.END}",
        (
            "\t3) On a shell command, type this for uploading the content to GitHub:\n"
            f"{bc.DIM}"
            "\t\t$ git add --all\n"
            '\t\t$ git commit -m "[MIG] {addon}: Migration to {to_branch}"\n'
            "\t\t$ git push {fork} {mig_branch} --set-upstream"
            f"{bc.END}"
        ),
        "\t4) Create the PR against {upstream_org}/{repo_name}:",
        f"\t\t=> {bc.BOLD}" "{new_pr_url}" f"{bc.END}",
    ]
)
BLACKLIST_TIPS = "\n".join(
    [
        f"\n{bc.BOLD}{bc.OKCYAN}The next steps are:{bc.END}",
        (
            "\t1) On a shell command, type this for uploading the content to GitHub:\n"
            f"{bc.DIM}"
            "\t\t$ git push {fork} {mig_branch} --set-upstream"
            f"{bc.END}"
        ),
        "\t2) Create the PR against {upstream_org}/{repo_name}:",
        f"\t\t=> {bc.BOLD}" "{new_pr_url}" f"{bc.END}",
    ]
)


class MigrateAddon:
    def __init__(
        self,
        repo,
        upstream_org,
        repo_name,
        from_branch,
        to_branch,
        fork,
        user_org,
        addon,
        storage,
        cache=None,
        verbose=False,
        non_interactive=False,
    ):
        self.repo = repo
        self.upstream_org = upstream_org
        self.repo_name = repo_name
        self.from_branch = from_branch
        self.to_branch = to_branch
        self.fork = fork
        self.user_org = user_org
        self.addon = addon
        self.storage = storage
        self.cache = cache
        self.mig_branch = g.Branch(
            repo, MIG_BRANCH_NAME.format(branch=to_branch.name[:4], addon=addon)
        )
        self.verbose = verbose
        self.non_interactive = non_interactive

    def run(self):
        blacklisted = self.storage.is_addon_blacklisted()
        if blacklisted:
            print(
                f"{bc.DIM}Migration of {bc.BOLD}{self.addon}{bc.END} "
                f"{bc.DIM}to {self.to_branch.name} "
                f"blacklisted ({blacklisted}){bc.ENDD}"
            )
            return
        if self.non_interactive:
            # Exit with an error code if the addon is eligible for a migration
            raise SystemExit(1)
        confirm = (
            f"Migrate {bc.BOLD}{self.addon}{bc.END} "
            f"from {bc.BOLD}{self.from_branch.name}{bc.END} "
            f"to {bc.BOLD}{self.to_branch.name}{bc.END}?"
        )
        if not click.confirm(confirm):
            self.storage.blacklist_addon(confirm=True)
            if not self.storage.dirty:
                return
        # Check if a migration PR already exists
        # TODO
        if not self.fork:
            raise click.UsageError("Please set the '--fork' option")
        if self.repo.untracked_files:
            raise click.ClickException("Untracked files detected, abort")
        self._checkout_base_branch()
        if self._create_mig_branch():
            # Case where the addon shouldn't be ported (blacklisted)
            if self.storage.dirty:
                self.storage.commit()
                self._print_tips(blacklisted=True)
                return
            with tempfile.TemporaryDirectory() as patches_dir:
                self._generate_patches(patches_dir)
                self._apply_patches(patches_dir)
            g.run_pre_commit(self.repo, self.addon)
        # Check if the addon has commits that update neighboring addons to
        # make it work properly
        PortAddonPullRequest(
            self.repo,
            self.upstream_org,
            self.repo_name,
            self.from_branch,
            self.mig_branch,
            self.fork,
            self.user_org,
            self.addon,
            self.storage,
            self.cache,
            self.verbose,
            create_branch=False,
            push_branch=False,
        ).run()
        self._print_tips()

    def _checkout_base_branch(self):
        # Ensure to not start to work from a working branch
        if self.to_branch.name in self.repo.heads:
            self.repo.heads[self.to_branch.name].checkout()
        else:
            self.repo.git.checkout(
                "--no-track", "-b", self.to_branch.name, self.to_branch.ref()
            )

    def _create_mig_branch(self):
        create_branch = True
        if self.mig_branch.name in self.repo.heads:
            confirm = (
                f"Branch {bc.BOLD}{self.mig_branch.name}{bc.END} already exists, "
                "recreate it?\n(⚠️  you will lose the existing branch)"
            )
            if click.confirm(confirm):
                self.repo.delete_head(self.mig_branch.name, "-f")
            else:
                create_branch = False
        if create_branch:
            # Create branch
            print(
                f"\tCreate branch {bc.BOLD}{self.mig_branch.name}{bc.END} "
                f"from {self.to_branch.ref()}..."
            )
            self.repo.git.checkout(
                "--no-track", "-b", self.mig_branch.name, self.to_branch.ref()
            )
        return create_branch

    def _generate_patches(self, patches_dir):
        print("\tGenerate patches...")
        self.repo.git.format_patch(
            "--keep-subject",
            "-o",
            patches_dir,
            f"{self.to_branch.ref()}..{self.from_branch.ref()}",
            "--",
            self.addon,
        )

    def _apply_patches(self, patches_dir):
        patches = [
            os.path.join(patches_dir, f) for f in sorted(os.listdir(patches_dir))
        ]
        # Apply patches with git-am
        print(f"\tApply {len(patches)} patches...")
        self.repo.git.am("-3", "--keep", *patches)
        print(
            f"\t\tCommits history of {bc.BOLD}{self.addon}{bc.END} "
            f"has been migrated."
        )

    def _print_tips(self, blacklisted=False):
        mig_tasks_url = MIG_TASKS_URL.format(branch=self.to_branch.name)
        pr_title_encoded = urllib.parse.quote(
            MIG_NEW_PR_TITLE.format(to_branch=self.to_branch.name[:4], addon=self.addon)
        )
        new_pr_url = MIG_NEW_PR_URL.format(
            upstream_org=self.upstream_org,
            repo_name=self.repo_name,
            to_branch=self.to_branch.name,
            user_org=self.user_org,
            mig_branch=self.mig_branch.name,
            title=pr_title_encoded,
        )
        if blacklisted:
            tips = BLACKLIST_TIPS.format(
                upstream_org=self.upstream_org,
                repo_name=self.repo_name,
                fork=self.fork,
                mig_branch=self.mig_branch.name,
                new_pr_url=new_pr_url,
            )
            print(tips)
            return
        tips = MIG_TIPS.format(
            upstream_org=self.upstream_org,
            repo_name=self.repo_name,
            addon=self.addon,
            to_branch=self.to_branch.name,
            fork=self.fork,
            mig_branch=self.mig_branch.name,
            mig_tasks_url=mig_tasks_url,
            new_pr_url=new_pr_url,
        )
        print(tips)



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import json
import os
from collections import defaultdict

import click

from . import git as g, misc


class InputStorage:
    """Store the user inputs related to an addon.

    If commits/pull requests of an addon may be ported, some of them could be
    false-positive and as such the user can choose to not port them.
    This class will help to store these informations so the tool won't list
    anymore these commits the next time we perform an analysis on the same addon.

    Technically the data are stored in one JSON file per addon in a hidden
    folder at the root of the repository. E.g:

        {
          "no_migration": "included in standard"
        }

    Or:

        {
          "pull_requests": {
            490: "lint changes"
          }
        }
    """

    storage_dirname = ".oca/oca-port"

    def __init__(self, to_branch, addon):
        self.to_branch = to_branch
        self.repo = self.to_branch.repo
        self.root_path = self.repo.working_dir
        self.addon = addon
        self._data = self._get_data()
        self.dirty = False

    def _get_data(self):
        """Return the data of the current repository.

        If a JSON file is found, return its content, otherwise return an empty
        dictionary.
        """
        try:
            # Read the JSON file from 'to_branch'
            tree = self.repo.commit(self.to_branch.ref()).tree
            blob = tree / self.storage_dirname / "blacklist" / f"{self.addon}.json"
            content = blob.data_stream.read().decode()
            return json.loads(content, object_hook=misc.defaultdict_from_dict)
        except KeyError:
            nested_dict = lambda: defaultdict(nested_dict)  # noqa
            return nested_dict()

    def save(self):
        """Store the data at the root of the current repository."""
        if not self._data or not self.dirty:
            return False
        file_path = self._get_file_path()
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as file_:
            json.dump(self._data, file_, indent=2)
        return True

    def _get_file_path(self):
        return os.path.join(
            self.root_path, self.storage_dirname, "blacklist", f"{self.addon}.json"
        )

    def is_pr_blacklisted(self, pr_ref):
        pr_ref = str(pr_ref or "orphaned_commits")
        return self._data.get("pull_requests", {}).get(pr_ref, False)

    def blacklist_pr(self, pr_ref, confirm=False, reason=None):
        if confirm and not click.confirm("\tBlacklist this PR?"):
            return
        if not reason:
            reason = click.prompt("\tReason", type=str)
        pr_ref = str(pr_ref or "orphaned_commits")
        self._data["pull_requests"][pr_ref] = reason or "Unknown"
        self.dirty = True

    def is_addon_blacklisted(self):
        entry = self._data.get("no_migration")
        if entry:
            return entry
        return False

    def blacklist_addon(self, confirm=False, reason=None):
        if confirm and not click.confirm("\tBlacklist this module?"):
            return
        if not reason:
            reason = click.prompt("Reason", type=str)
        self._data["no_migration"] = reason or "Unknown"
        self.dirty = True

    def commit(self):
        """Commit all files contained in the storage directory."""
        if not self.save():
            return
        changed_paths = g.get_changed_paths(self.repo)
        all_in_storage = all(
            path.startswith(self.storage_dirname) for path in changed_paths
        )
        if self.repo.is_dirty() and not all_in_storage:
            raise click.ClickException(
                "changes not committed detected in this repository."
            )
        # Ensure to be on a dedicated branch
        if self.repo.active_branch.name == self.to_branch.name:
            raise click.ClickException(
                "performing commit on upstream branch is not allowed."
            )
        # Commit all changes under ./.oca-port
        self.repo.index.add(self.storage_dirname)
        if self.repo.is_dirty():
            g.run_pre_commit(self.repo, self.addon, commit=False, hook="prettier")
            self.repo.index.commit(f"oca-port: store '{self.addon}' data")
            self.dirty = False
